"""
Base class for authenticated API calls used by Entity, Content and Upload
"""

import configparser
import hashlib
import os
import sys
import threading
import time
import xml.etree.ElementTree
from enum import Enum
import requests

import pyPreservica

CHUNK_SIZE = 1024 * 2

NS_XIPV6 = "http://preservica.com/XIP/v6.0"
NS_XIPV61 = "http://preservica.com/XIP/v6.1"
NS_XIPV62 = "http://preservica.com/XIP/v6.2"

NS_ENTITY = "http://preservica.com/EntityAPI/v6.0"
NS_ENTITY61 = "http://preservica.com/EntityAPI/v6.1"
NS_ENTITY62 = "http://preservica.com/EntityAPI/v6.2"

namespace = {'xip': NS_XIPV6, 'v6': NS_ENTITY}
HEADER_TOKEN = "Preservica-Access-Token"

IO_PATH = "information-objects"
SO_PATH = "structural-objects"
CO_PATH = "content-objects"

HASH_BLOCKSIZE = 65536


class FileHash:
    """
    A wrapper around the hashlib hash algorithms that allows an entire file to
    be hashed in a chunked manner.
    """

    def __init__(self, algorithm):
        self.algorithm = algorithm

    def get_algorithm(self):
        return self.algorithm

    def __call__(self, file):
        hash_algorithm = self.algorithm()
        with open(file, 'rb') as afile:
            buf = afile.read(HASH_BLOCKSIZE)
            while len(buf) > 0:
                hash_algorithm.update(buf)
                buf = afile.read(HASH_BLOCKSIZE)
        return hash_algorithm.hexdigest()


def only_assets(entity):
    return bool(entity.entity_type is EntityType.ASSET)


def only_folders(entity):
    return bool(entity.entity_type is EntityType.FOLDER)


class PagedSet:
    """
    Class to represent a page of results
    The results object contains the list of objects of interest
    """

    def __init__(self, results, has_more, total, next_page):
        self.results = results
        self.has_more = bool(has_more)
        self.total = int(total)
        self.next_page = next_page

    def __str__(self):
        return self.results.__str__()

    def get_results(self):
        return self.results

    def get_total(self):
        return self.total

    def has_more_pages(self):
        return self.has_more


class Sha1FixityCallBack:
    def __call__(self, filename, full_path):
        sha = FileHash(hashlib.sha1)
        return "SHA1", sha(full_path)


class Sha256FixityCallBack:
    def __call__(self, filename, full_path):
        sha = FileHash(hashlib.sha256)
        return "SHA256", sha(full_path)


class Sha512FixityCallBack:
    def __call__(self, filename, full_path):
        sha = FileHash(hashlib.sha512)
        return "SHA512", sha(full_path)


class UploadProgressCallback:
    """
    Default implementation of a callback class to show upload progress of a file
    """

    def __init__(self, filename):
        self._filename = filename
        self._size = float(os.path.getsize(filename))
        self._seen_so_far = 0
        self._lock = threading.Lock()

    def __call__(self, bytes_amount):
        with self._lock:
            self._seen_so_far += bytes_amount
            percentage = (self._seen_so_far / self._size) * 100
            sys.stdout.write("\r%s  %s / %s  (%.2f%%)" % (self._filename, self._seen_so_far, self._size, percentage))
            sys.stdout.flush()


class IntegrityCheck:
    """
    Class to hold information about completed integrity checks
    """

    def __init__(self, check_type, success, date, adapter, fixed, reason):
        self.check_type = check_type
        self.success = bool(success)
        self.date = date
        self.adapter = adapter
        self.fixed = bool(fixed)
        self.reason = reason

    def __str__(self):
        return f"Type:\t\t\t{self.check_type}\n" \
               f"Success:\t\t\t{self.success}\n" \
               f"Date:\t{self.date}\n" \
               f"Storage Adapter:\t{self.adapter}\n"

    def __repr__(self):
        return self.__str__()

    def get_adapter(self):
        return self.adapter

    def get_success(self):
        return self.success


class Representation:
    """
        Class to represent the Representation Object in the Preservica data model
    """

    def __init__(self, asset, rep_type, name, url):
        self.asset = asset
        self.rep_type = rep_type
        self.name = name
        self.url = url

    def __str__(self):
        return f"Type:\t\t\t{self.rep_type}\n" \
               f"Name:\t\t\t{self.name}\n" \
               f"URL:\t{self.url}"

    def __repr__(self):
        return self.__str__()


class Bitstream:
    """
        Class to represent the Bitstream Object or digital file in the Preservica data model
    """

    def __init__(self, filename, length, fixity, content_url):
        self.filename = filename
        self.length = int(length)
        self.fixity = fixity
        self.content_url = content_url

    def __str__(self):
        return f"Filename:\t\t\t{self.filename}\n" \
               f"FileSize:\t\t\t{self.length}\n" \
               f"Content:\t{self.content_url}\n" \
               f"Fixity:\t{self.fixity}"

    def __repr__(self):
        return self.__str__()


class Generation:
    """
         Class to represent the Generation Object in the Preservica data model
     """

    def __init__(self, original, active, format_group, effective_date, bitstreams):
        self.original = original
        self.active = active
        self.content_object = None
        self.format_group = format_group
        self.effective_date = effective_date
        self.bitstreams = bitstreams

    def __str__(self):
        return f"Active:\t\t\t{self.active}\n" \
               f"Original:\t\t\t{self.original}\n" \
               f"Format_group:\t{self.format_group}"

    def __repr__(self):
        return self.__str__()


class Entity:
    """
        Base Class of Assets, Folders and Content Objects
    """

    def __init__(self, reference, title, description, security_tag, parent, metadata):
        self.reference = reference
        self.title = title
        self.description = description
        self.security_tag = security_tag
        self.parent = parent
        self.metadata = metadata
        self.entity_type = None
        self.path = None
        self.tag = None

    def __str__(self):
        return f"Ref:\t\t\t{self.reference}\n" \
               f"Title:\t\t\t{self.title}\n" \
               f"Description:\t{self.description}\n" \
               f"Security Tag:\t{self.security_tag}\n" \
               f"Parent:\t\t\t{self.parent}\n\n"

    def __repr__(self):
        return self.__str__()


class Folder(Entity):
    """
       Class to represent the Structural Object or Folder in the Preservica data model
    """

    def __init__(self, reference, title, description, security_tag, parent, metadata):
        super().__init__(reference, title, description, security_tag, parent, metadata)
        self.entity_type = EntityType.FOLDER
        self.path = SO_PATH
        self.tag = "StructuralObject"


class Asset(Entity):
    """
        Class to represent the Information Object or Asset in the Preservica data model
    """

    def __init__(self, reference, title, description, security_tag, parent, metadata):
        super().__init__(reference, title, description, security_tag, parent, metadata)
        self.entity_type = EntityType.ASSET
        self.path = IO_PATH
        self.tag = "InformationObject"


class ContentObject(Entity):
    """
       Class to represent the Content Object in the Preservica data model
    """

    def __init__(self, reference, title, description, security_tag, parent, metadata):
        super().__init__(reference, title, description, security_tag, parent, metadata)
        self.entity_type = EntityType.CONTENT_OBJECT
        self.representation_type = None
        self.asset = None
        self.path = CO_PATH
        self.tag = "ContentObject"


def content_api_identifier_to_type(ref):
    ref = ref.replace('sdb:', '')
    parts = ref.split("|")
    return tuple((EntityType(parts[0]), parts[1]))


def entity_from_string(xml_data):
    entity_response = xml.etree.ElementTree.fromstring(xml_data)
    reference = entity_response.find('.//{*}Ref')
    title = entity_response.find('.//{*}Title')
    security_tag = entity_response.find('.//{*}SecurityTag')
    description = entity_response.find('.//{*}Description')
    parent = entity_response.find('.//{*}Parent')
    if hasattr(parent, 'text'):
        parent = parent.text
    else:
        parent = None

    fragments = entity_response.findall('.//{*}Metadata/{*}Fragment')
    metadata = {}
    for fragment in fragments:
        metadata[fragment.text] = fragment.attrib['schema']

    return {'reference': reference.text, 'title': title.text if hasattr(title, 'text') else None,
            'description': description.text if hasattr(description, 'text') else None,
            'security_tag': security_tag.text, 'parent': parent, 'metadata': metadata}


class Thumbnail(Enum):
    SMALL = "small"
    MEDIUM = "medium"
    LARGE = "large"


class EntityType(Enum):
    ASSET = "IO"
    FOLDER = "SO"
    CONTENT_OBJECT = "CO"


class AuthenticatedAPI:
    """
    Base class for authenticated calls which need an access token
    """

    def __version_number__(self):
        headers = {HEADER_TOKEN: self.token}
        request = requests.get(f'https://{self.server}/api/entity/versiondetails/version', headers=headers)
        if request.status_code == requests.codes.ok:
            xml_ = str(request.content)
            version = xml_[xml_.find("<CurrentVersion>") + len("<CurrentVersion>"):xml_.find("</CurrentVersion>")]
            version_numbers = version.split(".")
            self.major_version = int(version_numbers[0])
            self.minor_version = int(version_numbers[1])
            self.patch_version = int(version_numbers[2])
            return version
        if request.status_code == requests.codes.unauthorized:
            self.token = self.__token__()
            return self.__version_number__()
        raise RuntimeError(request.status_code, "version number failed")

    def __str__(self):
        return f"pyPreservica version: {pyPreservica.__version__}  (Preservica 6.2 Compatible) \n" \
               f"Connected to: {self.server} Preservica version: {self.version} as {self.username}\n"

    def __repr__(self):
        return self.__str__()

    def save_config(self):
        config = configparser.ConfigParser()
        config['credentials'] = {'username': self.username, 'password': self.password, 'tenant': self.tenant,
                                 'server': self.server}
        with open('credentials.properties', 'wt') as configfile:
            config.write(configfile)

    def manager_token(self, username, password):
        data = {'username': username, 'password': password, 'tenant': self.tenant}
        response = requests.post(f'https://{self.server}/api/accesstoken/login', data=data)
        if response.status_code == requests.codes.ok:
            return response.json()['token']
        raise RuntimeError(response.status_code, "Could not generate valid manager approval password")

    def __token__(self):

        if self.shared_secret is False:
            data = {'username': self.username, 'password': self.password, 'tenant': self.tenant}
            response = requests.post(f'https://{self.server}/api/accesstoken/login', data=data)
            if response.status_code == requests.codes.ok:
                return response.json()['token']
            else:
                raise RuntimeError(response.status_code,
                                   "Failed to create an password based authentication token. Check your credentials "
                                   "are correct")

        if self.shared_secret is True:
            endpoint = "api/accesstoken/acquire-external"
            timestamp = int(time.time())
            to_hash = f"preservica-external-auth{timestamp}{self.username}{self.password}"
            sha1 = hashlib.sha1()
            sha1.update(to_hash.encode(encoding='UTF-8'))
            data = {"username": self.username, "tenant": self.tenant, "timestamp": timestamp, "hash": sha1.hexdigest()}
            response = requests.post(f'https://{self.server}/{endpoint}', data=data)
            if response.status_code == requests.codes.ok:
                return response.json()['token']
            else:
                raise RuntimeError(response.status_code,
                                   "Failed to create an shared secret authentication token. Check your credentials "
                                   "are correct")

    def __init__(self, username=None, password=None, tenant=None, server=None, use_shared_secret=False):
        config = configparser.ConfigParser()
        config.read('credentials.properties')

        self.shared_secret = bool(use_shared_secret)

        if not username:
            username = os.environ.get('PRESERVICA_USERNAME')
            if username is None:
                try:
                    username = config['credentials']['username']
                except KeyError:
                    pass
        if not username:
            print("No valid username found in method arguments, environment variables or credentials.properties file")
            raise SystemExit
        else:
            self.username = username

        if not password:
            password = os.environ.get('PRESERVICA_PASSWORD')
            if password is None:
                try:
                    password = config['credentials']['password']
                except KeyError:
                    pass
        if not password:
            print("No valid password found in method arguments, environment variables or credentials.properties file")
            raise SystemExit
        else:
            self.password = password

        if not tenant:
            tenant = os.environ.get('PRESERVICA_TENANT')
            if tenant is None:
                try:
                    tenant = config['credentials']['tenant']
                except KeyError:
                    pass
        if not tenant:
            print("No valid tenant found in method arguments, environment variables or credentials.properties file")
            raise SystemExit
        else:
            self.tenant = tenant

        if not server:
            server = os.environ.get('PRESERVICA_SERVER')
            if server is None:
                try:
                    server = config['credentials']['server']
                except KeyError:
                    pass
        if not server:
            print("No valid server found in method arguments, environment variables or credentials.properties file")
            raise SystemExit
        else:
            self.server = server

        self.token = self.__token__()
        self.version = self.__version_number__()



import uuid
from xml.dom import minidom
from xml.etree import ElementTree

from pyPreservica.common import *


def prettify(elem):
    """Return a pretty-printed XML string for the Element.
    """
    rough_string = ElementTree.tostring(elem, 'utf-8')
    re_parsed = minidom.parseString(rough_string)
    return re_parsed.toprettyxml(indent="  ")


class WorkflowContext:
    def __init__(self, workflow_id, workflow_name):
        self.workflow_id = workflow_id
        self.workflow_name = workflow_name

    def __str__(self):
        return f"Workflow ID:\t\t\t{self.workflow_id}\n" \
               f"Workflow Name:\t\t\t{self.workflow_name}\n"

    def __repr__(self):
        return self.__str__()


class WorkflowAPI(AuthenticatedAPI):
    """
              A client library for the Preservica Workflow API
              https://preview.preservica.com/sdb/rest/workflow/documentation.html

    """

    def __init__(self, username=None, password=None, tenant=None, server=None, use_shared_secret=False):
        super().__init__(username, password, tenant, server, use_shared_secret)
        self.base_url = "sdb/rest/workflow"

    def get_workflow_contexts(self, definition: str):
        """
        Return a list of Workflow Contexts which have the same Workflow Definition ID

        :param definition: The Workflow Definition ID

        """
        headers = {HEADER_TOKEN: self.token}
        payload = {"workflowDefinitionId": definition}
        workflow_contexts = list()
        request = requests.get(f'https://{self.server}/{self.base_url}/contexts', headers=headers, params=payload)
        if request.status_code == requests.codes.ok:
            xml_response = str(request.content.decode('utf-8'))
            entity_response = xml.etree.ElementTree.fromstring(xml_response)
            contexts = entity_response.findall(".//{*}WorkflowContext")
            for context in contexts:
                wrkfl_id = context.find(".//{*}Id").text
                name = context.find(".//{*}Name").text
                workflow_context = WorkflowContext(wrkfl_id, name)
                workflow_contexts.append(workflow_context)
            return workflow_contexts
        elif request.status_code == requests.codes.unauthorized:
            self.token = self.__token__()
            return self.get_workflow_contexts(definition)
        else:
            raise RuntimeError(request.status_code, "get_workflow_contexts")

    def start_workflow_instance(self, workflow_context, **kwargs):
        """

        Start a workflow context

        Returns a Correlation Id which is used to monitor the workflow progress

        :param workflow_context: The workflow context to start
        :param kwargs:           Key/Values to pass to the workflow instance

        """
        headers = {HEADER_TOKEN: self.token, 'Content-Type': 'application/xml;charset=UTF-8'}

        correlation_id = str(uuid.uuid4())

        request_payload = xml.etree.ElementTree.Element('StartWorkflowRequest ',
                                                        {"xmlns": "http://workflow.preservica.com"})
        xml.etree.ElementTree.SubElement(request_payload, "WorkflowContextId").text = workflow_context.workflow_id
        xml.etree.ElementTree.SubElement(request_payload, "WorkflowContextName").text = workflow_context.workflow_name

        for key, value in kwargs.items():
            parameter = xml.etree.ElementTree.SubElement(request_payload, "Parameter")
            xml.etree.ElementTree.SubElement(parameter, "Key").text = key
            xml.etree.ElementTree.SubElement(parameter, "Value").text = value

        xml.etree.ElementTree.SubElement(request_payload, "CorrelationId").text = correlation_id

        xml_request = xml.etree.ElementTree.tostring(request_payload, encoding='utf-8')
        request = requests.post(f'https://{self.server}/{self.base_url}/instances', headers=headers, data=xml_request)
        if request.status_code == requests.codes.created:
            return correlation_id
        if request.status_code == requests.codes.unauthorized:
            self.token = self.__token__()
            return self.start_workflow_instance(workflow_context, **kwargs)
        else:
            raise RuntimeError(request.status_code, "start_workflow_instance failed")



import xml.etree.ElementTree
from pyPreservica.common import *


class RetentionPolicy:
    def __init__(self, name, ref):
        self.name = name
        self.ref = ref

    def __str__(self):
        return f"Ref:\t\t\t{self.ref}\n" \
               f"Name:\t\t\t{self.name}\n"

    def __repr__(self):
        return self.__str__()


class RetentionAPI(AuthenticatedAPI):

    def __init__(self, username=None, password=None, tenant=None, server=None, use_shared_secret=False):
        super().__init__(username, password, tenant, server, use_shared_secret)
        if self.major_version < 7 and self.minor_version < 2:
            raise RuntimeError("Retention API is only available when connected to a v6.2 System")

    def policy(self, reference: str) -> RetentionPolicy:
        """
           Return a retention policy by its reference
        """
        headers = {HEADER_TOKEN: self.token, 'Content-Type': 'application/xml;charset=UTF-8'}
        request = requests.get(f'https://{self.server}/api/entity/retention-policies/{reference}', headers=headers)
        if request.status_code == requests.codes.ok:
            xml_response = str(request.content.decode('utf-8'))
            entity_response = xml.etree.ElementTree.fromstring(xml_response)
            ref = entity_response.find('.//{*}RetentionPolicy/{*}Ref').text
            assert ref == reference
            name = entity_response.find('.//{*}RetentionPolicy/{*}Name').text
            rp = RetentionPolicy(name, ref)
            description = entity_response.find('.//{*}RetentionPolicy/{*}Description').text
            rp.description = description
            security_tag = entity_response.find('.//{*}RetentionPolicy/{*}SecurityTag').text
            rp.security_tag = security_tag
            start_date_field = entity_response.find('.//{*}RetentionPolicy/{*}StartDateField').text
            rp.start_date_field = start_date_field
            period = entity_response.find('.//{*}RetentionPolicy/{*}Period').text
            rp.period = period
            period_unit = entity_response.find('.//{*}RetentionPolicy/{*}PeriodUnit').text
            rp.period_unit = period_unit
            expiry_action = entity_response.find('.//{*}RetentionPolicy/{*}ExpiryAction').text
            rp.expiry_action = expiry_action
            restriction = entity_response.find('.//{*}RetentionPolicy/{*}Restriction')
            if restriction is not None:
                rp.restriction = restriction.text
            else:
                rp.restriction = None
            assignable = entity_response.find('.//{*}RetentionPolicy/{*}Assignable')
            rp.assignable = bool(assignable.text == "true")
            return rp
        elif request.status_code == requests.codes.unauthorized:
            self.token = self.__token__()
            return self.policy(reference)
        else:
            raise RuntimeError(request.status_code, "policy failed")

    def create_policy(self, name: str, description: str = None, security_tag: str = "open"):
        pass

    def delete_policy(self, reference: str):
        pass

    def policies(self):
        """
        Return a list of all retention policies
        """
        headers = {HEADER_TOKEN: self.token, 'Content-Type': 'application/xml;charset=UTF-8'}
        data = {'start': str(0), 'max': "250"}
        request = requests.get(f'https://{self.server}/api/entity/retention-policies', data=data, headers=headers)
        if request.status_code == requests.codes.ok:
            xml_response = str(request.content.decode('utf-8'))
            entity_response = xml.etree.ElementTree.fromstring(xml_response)
            result = set()
            for assignment in entity_response.findall('.//{*}RetentionPolicy'):
                ref = assignment.attrib['ref']
                name = assignment.attrib['name']
                rp = RetentionPolicy(name, ref)
                result.add(rp)
            return result
        elif request.status_code == requests.codes.unauthorized:
            self.token = self.__token__()
            return self.policies()
        else:
            raise RuntimeError(request.status_code, "policies failed")

    def add_assignments(self, entity: Entity, policy: RetentionPolicy):
        pass

    def remove_assignments(self, entity: Entity, policy: RetentionPolicy):
        pass

    def assignments(self, entity: Entity):
        """
          Return a list of retention policies for an entity.
        """
        headers = {HEADER_TOKEN: self.token, 'Content-Type': 'application/xml;charset=UTF-8'}
        request = requests.get(
            f'https://{self.server}/api/entity/{entity.path}/{entity.reference}/retention-assignments',
            headers=headers)
        if request.status_code == requests.codes.ok:
            xml_response = str(request.content.decode('utf-8'))
            entity_response = xml.etree.ElementTree.fromstring(xml_response)
            result = set()
            for assignment in entity_response.findall('.//{*}RetentionAssignment'):
                entity_ref = assignment.find('./{*}Entity').text
                assert entity_ref == entity.reference
                policy = assignment.find('./{*}RetentionPolicy').text
                date = assignment.find('./{*}StartDate').text
                expired = bool(assignment.find('./{*}Expired').text == 'true')
                api_id = assignment.find('./{*}ApiId').text
                result.add(self.policy(policy))
            return result
        elif request.status_code == requests.codes.unauthorized:
            self.token = self.__token__()
            return self.assignments(entity)
        else:
            raise RuntimeError(request.status_code, "assignments failed")



from .common import *
from .contentAPI import ContentAPI
from .entityAPI import EntityAPI
from .uploadAPI import UploadAPI, simple_asset_package, complex_asset_package, cvs_to_xsd, cvs_to_xml, \
                    cvs_to_cmis_xslt, csv_to_search_xml
from .workflowAPI import WorkflowAPI
from .retentionAPI import RetentionAPI

__author__ = "James Carr (drjamescarr@gmail.com)"

# Version of the Preservica API package
__version__ = "0.9.0"

__license__ = "Apache License Version 2.0"
